from azure_explorer.config import Config
from azure_explorer.managers import (
    ContainerManager,
    ServiceBusManager,
    StorageAccountManager,
    SubscriptionManager,
    TenantManager,
)


def get_tenant_manager() -> TenantManager:
    return TenantManager(
        credential=Config.credential,
    )


def get_subscription_manager(subscription_id: str) -> SubscriptionManager:
    return SubscriptionManager(
        subscription_id=subscription_id,
        credential=Config.credential,
    )


def get_storage_account_manager(
    storage_account_name: str,
) -> StorageAccountManager:
    return StorageAccountManager(storage_account_name, credential=Config.credential)


def get_container_manager(
    storage_account_name: str,
    container_name: str,
    is_data_lake: bool = False,
) -> ContainerManager:
    return ContainerManager.create(
        storage_account_name, container_name, is_data_lake, credential=Config.credential
    )


def get_service_bus_manager(
    service_bus_name: str,
) -> ServiceBusManager:
    return ServiceBusManager(
        service_bus_name=service_bus_name,
        credential=Config.credential,
    )


def read_file(
    storage_account_name: str,
    container_name: str,
    path: str,
) -> bytes:
    container_manager = get_container_manager(
        storage_account_name, container_name, is_data_lake=False
    )
    return container_manager.read_file(path)



from azure.core.credentials import TokenCredential
from azure.mgmt.resource import ResourceManagementClient
from azure.mgmt.servicebus import ServiceBusManagementClient
from azure.mgmt.storage import StorageManagementClient

from azure_explorer.managers.base import Manager
from azure_explorer.managers.models import (
    ServiceBusProperties,
    StorageAccountProperties,
)
from azure_explorer.managers.service_bus import ServiceBusManager
from azure_explorer.managers.storage_account import StorageAccountManager


class SubscriptionManager(Manager):
    def __init__(self, subscription_id: str, credential: TokenCredential):
        self.subscription_id = subscription_id
        self.resource_mgtm_client = ResourceManagementClient(
            credential,
            subscription_id,
        )
        self.storage_mgtm_client = StorageManagementClient(
            credential,
            subscription_id,
        )
        self.servicebus_mgtm_client = ServiceBusManagementClient(
            credential,
            subscription_id,
        )
        self.credential = credential

    def list_storage_accounts(self) -> list[StorageAccountProperties]:
        items = []
        for storage_account in self.storage_mgtm_client.storage_accounts.list():
            item = StorageAccountProperties(
                storage_account.name,
                is_data_lake=storage_account.is_hns_enabled or False,
            )
            items.append(item)

        return items

    def list_storage_account_ids(self) -> list[str]:
        return [item.id for item in self.list_storage_accounts()]

    def list_storage_account_names(self) -> list[str]:
        return [item.name for item in self.list_storage_accounts()]

    def get_storage_account_manager(
        self, storage_account_name: str, is_data_lake: bool
    ):
        return StorageAccountManager(
            storage_account_name, self.credential, is_data_lake
        )

    def list_service_buses(self) -> list[ServiceBusProperties]:
        items = []
        for namespace in self.servicebus_mgtm_client.namespaces.list():
            item = ServiceBusProperties(
                namespace.name,
            )
            items.append(item)

        return items

    def list_service_bus_ids(self) -> list[str]:
        return [item.id for item in self.list_service_buses()]

    def list_service_bus_names(self) -> list[str]:
        return [item.name for item in self.list_service_buses()]

    def get_service_bus_manager(self, service_bus_name: str):
        return ServiceBusManager(service_bus_name, self.credential)

    def __repr__(self) -> str:
        return f"SubscriptionManager(subscription_id='{self.subscription_id}')"



from azure.core.credentials import TokenCredential
from azure.servicebus.management import ServiceBusAdministrationClient

from azure_explorer.managers.base import Manager
from azure_explorer.managers.models import TopicSubscriptionProperties
from azure_explorer.managers.topic_subscription import TopicSubscriptionManager


class TopicManager(Manager):
    def __init__(
        self, service_bus_name: str, topic_name: str, credential: TokenCredential
    ):
        self.service_bus_name = service_bus_name
        self.credential = credential
        self.topic_name = topic_name
        self.service_bus_adm_client = ServiceBusAdministrationClient(
            fully_qualified_namespace=f"{service_bus_name}.servicebus.windows.net",
            credential=credential,
        )

    def check_access(self) -> bool:
        next(
            self.service_bus_adm_client.list_subscriptions_runtime_properties(
                self.topic_name
            )
        )

    def list_subscription(self) -> list[TopicSubscriptionProperties]:
        subscriptions = []
        for (
            sub_runtime_props
        ) in self.service_bus_adm_client.list_subscriptions_runtime_properties(
            self.topic_name
        ):
            subscription = TopicSubscriptionProperties(
                name=sub_runtime_props.name,
                num_active_messages=sub_runtime_props.active_message_count,
                num_deadletter_messages=sub_runtime_props.dead_letter_message_count,
            )
            subscriptions.append(subscription)
        return subscriptions

    def list_subscription_names(self) -> list[str]:
        return [sub.name for sub in self.list_subscription()]

    def get_subscription_manager(
        self,
        subscription_name: str,
    ) -> TopicSubscriptionManager:
        return TopicSubscriptionManager(
            self.service_bus_name,
            self.topic_name,
            subscription_name,
            self.credential,
        )



class AccessError(Exception):
    pass


class ConnectionError(AccessError):
    def __init__(self, url: str):
        self.url = url
        super().__init__(
            f"Unable to connect to '{url}'. Are you working in a VNET, then remember "
            "to add a private endpoint for this resource."
        )


class AuthorizationError(AccessError):
    pass



from azure.core.credentials import TokenCredential
from azure.mgmt.subscription import SubscriptionClient

from azure_explorer.managers.base import Manager
from azure_explorer.managers.models import SubscriptionProperties
from azure_explorer.managers.subscription import SubscriptionManager


class TenantManager(Manager):
    def __init__(self, credential: TokenCredential):
        self.sub_client = SubscriptionClient(credential)
        self.credential = credential

    def check_access(self) -> bool:
        next(self.sub_client.subscriptions.list())

    def list_subscriptions(self) -> list[SubscriptionProperties]:
        items = []
        for sub in self.sub_client.subscriptions.list():
            item = SubscriptionProperties(
                sub.subscription_id,
                sub.display_name,
            )
            items.append(item)
        return items

    def list_subscription_ids(self) -> list[str]:
        return [item.id for item in self.list_subscriptions()]

    def list_subscription_names(self) -> list[str]:
        return [item.name for item in self.list_subscriptions()]

    def get_subscription_manager(self, subscription_id: str):
        return SubscriptionManager(subscription_id, self.credential)

    def __repr__(self) -> str:
        return "TenantManager()"



import logging

from textual.binding import Binding
from textual.widget import Widget

logging.basicConfig(
    filename="azure_explorer.log",
    level=logging.DEBUG,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    filemode="a",
)


class Window(Widget):

    BINDINGS = [
        Binding("escape", "go_to_parent_window", "Go Back", show=True, priority=True),
    ]

    def __init__(self):
        """Widget representing a window in a stack of windows,
        meaning it is possible to (1) go back to the parent window
        or (2) open a child window.
        """
        self.parent_window = None
        super().__init__()

    def check_valid(self) -> bool:
        return True

    def action_go_to_parent_window(self):
        if self.parent_window is not None:
            self.remove()
            self.app.mount(self.parent_window)

    def open_child_window(self, window: "Window"):
        try:
            window.check_valid()
        except Exception as exc:
            self.app.notify(str(exc), severity="error")
            return

        window.parent_window = self
        self.remove()
        self.app.mount(window)



from .base import Manager
from .container import ContainerManager
from .models import MessageProperties, SubscriptionProperties
from .service_bus import ServiceBusManager
from .storage_account import StorageAccountManager
from .subscription import SubscriptionManager
from .tenant import TenantManager
from .topic import TopicManager
from .topic_subscription import TopicSubscriptionManager



import socket

from azure_explorer.managers.exceptions import ConnectionError


def check_connection(url: str):
    try:
        socket.getaddrinfo(url, 443)
    except socket.gaierror:
        raise ConnectionError(url)



from azure.core.credentials import TokenCredential
from azure.core.exceptions import HttpResponseError
from azure.data.tables import TableServiceClient
from azure.storage.blob import BlobServiceClient

from azure_explorer.managers import utils
from azure_explorer.managers.base import Manager
from azure_explorer.managers.container import (
    BlobStorageContainerManager,
    DataLakeContainerManager,
)


class StorageAccountManager(Manager):
    def __init__(
        self, storage_account_name: str, credential: TokenCredential, is_data_lake: bool
    ):
        self.storage_account_name = storage_account_name
        self.credential = credential
        self.blob_endpoint = f"{storage_account_name}.blob.core.windows.net"
        self.table_endpoint = f"{storage_account_name}.table.core.windows.net"
        self.blob_service_client = BlobServiceClient(
            account_url=f"https://{self.blob_endpoint}",
            credential=credential,
        )
        self.table_service_client = TableServiceClient(
            endpoint=f"https://{self.table_endpoint}",
            credential=credential,
        )
        self.is_data_lake = is_data_lake

    def check_access(self, data_object_type: str) -> bool:
        if data_object_type == "container":
            utils.check_connection(self.blob_endpoint)
            try:
                next(self.blob_service_client.list_containers())
            except HttpResponseError as exc:
                raise exc
            except Exception as exc:
                raise exc
        elif data_object_type == "table":
            next(self.table_service_client.list_tables())
        else:
            raise ValueError(f"Unknown data object type: {data_object_type}")
        return True

    def get_container_manager(
        self, container_name: str
    ) -> DataLakeContainerManager | BlobStorageContainerManager:
        if self.is_data_lake:
            return DataLakeContainerManager(
                self.storage_account_name,
                container_name,
                self.credential,
            )
        else:
            return BlobStorageContainerManager(
                self.storage_account_name,
                container_name,
                self.credential,
            )

    def list_container_names(self) -> list[str]:
        return [
            container.name for container in self.blob_service_client.list_containers()
        ]

    def list_table_names(self) -> list[str]:
        return [table.name for table in self.table_service_client.list_tables()]

    def __repr__(self) -> str:
        return (
            "StorageAccountManager("
            f"storage_account_name='{self.storage_account_name}', "
            f"is_data_lake={self.is_data_lake})"
        )



from azure.core.credentials import TokenCredential
from azure.servicebus.management import ServiceBusAdministrationClient

from azure_explorer.managers import utils
from azure_explorer.managers.base import Manager
from azure_explorer.managers.models import TopicProperties
from azure_explorer.managers.topic import TopicManager


class ServiceBusManager(Manager):
    def __init__(self, service_bus_name: str, credential: TokenCredential):
        self.service_bus_name = service_bus_name
        self.credential = credential
        self.fully_qualified_namespace = f"{service_bus_name}.servicebus.windows.net"
        self.service_bus_adm_client = ServiceBusAdministrationClient(
            self.fully_qualified_namespace,
            credential,
        )

    def get_topic_manager(self, topic_name: str) -> TopicManager:
        return TopicManager(self.service_bus_name, topic_name, self.credential)

    def check_access(self) -> bool:
        utils.check_connection(self.fully_qualified_namespace)
        next(self.service_bus_adm_client.list_topics_runtime_properties())

    def list_topics(self) -> list[TopicProperties]:
        topics = []
        for (
            topic_runtime_props
        ) in self.service_bus_adm_client.list_topics_runtime_properties():
            topic = TopicProperties(
                name=topic_runtime_props.name,
            )
            topics.append(topic)
        return topics

    def list_topic_names(self) -> list[str]:
        return [topic.name for topic in self.list_topics()]

    def list_queue_names(self) -> list[str]:
        return [queue.name for queue in self.service_bus_adm_client.list_queues()]

    def __repr__(self) -> str:
        return f"ServiceBusManager(service_bus_name={self.service_bus_name})"



from io import BytesIO
from pathlib import Path
from typing import Iterator

import pytz
from azure.core.credentials import TokenCredential
from azure.storage.filedatalake import FileSystemClient, StorageStreamDownloader

from azure_explorer.managers.container.base import ContainerManager
from azure_explorer.managers.models import FileProperties, FolderProperties

UTC = pytz.utc


class DataLakeContainerManager(ContainerManager):
    def __init__(
        self,
        storage_account_name: str,
        container_name: str,
        credential: TokenCredential,
    ):
        self.storage_account_name = storage_account_name
        self.container_name = container_name
        self.credential = credential
        self.file_system_client = FileSystemClient(
            f"https://{storage_account_name}.dfs.core.windows.net",
            file_system_name=container_name,
            credential=credential,
        )

    def is_dir(self, folder: Path | str) -> bool:
        # checking if a folder exist is a bit tricky, but
        # we use the solution proposed here in
        # https://github.com/Azure/azure-sdk-for-python/issues/24814
        if folder == "":
            return True

        file_client = self.file_system_client.get_file_client(folder)
        if not file_client.exists():
            return False

        file_metadata = file_client.get_file_properties().metadata
        is_dir_blob = file_metadata.get("hdi_isfolder", "false").lower() == "true"
        if not is_dir_blob:
            return False

        return True

    def is_file(self, path: Path | str):
        if path == "":
            return False

        file_client = self.file_system_client.get_file_client(path)
        if not file_client.exists():
            return False

        file_metadata = file_client.get_file_properties().metadata
        is_dir_blob = file_metadata.get("hdi_isfolder", "false").lower() == "true"
        if is_dir_blob:
            return False

        return True

    def _iter_dir(
        self,
        folder: Path | str,
        recursive: bool,
    ) -> Iterator[FileProperties | FolderProperties]:
        for path in self.file_system_client.get_paths(folder, recursive):
            last_modified_time = path.last_modified.replace(tzinfo=UTC)
            created_time = path.creation_time.replace(tzinfo=UTC)

            if path.is_directory:
                if recursive:
                    continue
                yield FolderProperties(
                    path.name, None, last_modified_time, created_time
                )
            else:
                yield FileProperties(
                    path.name, path.content_length, last_modified_time, created_time
                )

    def _get_blob_stream_downloader(self, path: Path | str) -> StorageStreamDownloader:
        file_client = self.file_system_client.get_file_client(path)
        return file_client.download_file()

    def _write_blob_data(
        self, path: Path | str, data: bytes | BytesIO, overwrite: bool = False
    ):
        file_client = self.file_system_client.get_file_client(path)
        file_client.upload_data(data, overwrite=overwrite)

    def __repr__(self) -> str:
        return (
            "DataLakeContainerManager("
            f"storage_account_name='{self.storage_account_name}', "
            f"container_name='{self.container_name}')"
        )



from typing import Iterator

from textual.widget import Widget

from azure_explorer.managers import TenantManager
from azure_explorer.widgets.components.azure.subscription import ResourceTypeSelector
from azure_explorer.widgets.components.windows import Browser


class SubscriptionBrowser(Browser):
    COLUMN_NAMES = ["Name", "ID"]

    def __init__(self, tenant_manager: TenantManager):
        self.tenant_manager = tenant_manager
        super().__init__()

    def check_valid(self):
        self.tenant_manager.check_access()

    def iter_rows(self) -> Iterator[tuple[str, tuple]]:
        for subscription in self.tenant_manager.list_subscriptions():
            yield subscription.id, (subscription.name, subscription.id)

    def get_item_widget(self, id_: str) -> Widget:
        subscription_manager = self.tenant_manager.get_subscription_manager(
            subscription_id=id_,
        )
        return ResourceTypeSelector(subscription_manager)



from typing import Iterator

from azure.core.credentials import TokenCredential
from azure.servicebus import ServiceBusClient, ServiceBusSubQueue
from azure.servicebus.management import ServiceBusAdministrationClient

from azure_explorer.managers.base import Manager
from azure_explorer.managers.models import MessageProperties

PEEK_MESSAGE_BATCH_SIZE = 10


class TopicSubscriptionManager(Manager):
    def __init__(
        self,
        service_bus_name: str,
        topic_name: str,
        subscription_name: str,
        credential: TokenCredential,
    ):
        self.service_bus_name = service_bus_name
        self.credential = credential
        self.topic_name = topic_name
        self.subscription_name = subscription_name
        self.service_bus_client = ServiceBusClient(
            fully_qualified_namespace=f"{service_bus_name}.servicebus.windows.net",
            credential=credential,
        )
        self.service_bus_adm_client = ServiceBusAdministrationClient(
            fully_qualified_namespace=f"{service_bus_name}.servicebus.windows.net",
            credential=credential,
        )
        self.runtime_properties = (
            self.service_bus_adm_client.get_subscription_runtime_properties(
                self.topic_name, self.subscription_name
            )
        )

    def check_access(self) -> bool:
        receiver = self.service_bus_client.get_subscription_receiver(
            self.topic_name, self.subscription_name
        )
        receiver.peek_messages(max_message_count=1)

    def num_active_messages(self):
        return self.runtime_properties.active_message_count

    def num_deadletter_messages(self):
        return self.runtime_properties.dead_letter_message_count

    def _iter_sub_queue_messages(self, sub_queue) -> list[MessageProperties]:
        receiver = self.service_bus_client.get_subscription_receiver(
            self.topic_name, self.subscription_name, sub_queue=sub_queue
        )
        seq_num = 0
        # iterate in batches of PEEK_MESSAGE_BATCH_SIZE
        while True:
            empty_batch = True
            for msg in receiver.peek_messages(
                max_message_count=PEEK_MESSAGE_BATCH_SIZE,
                sequence_number=seq_num,
            ):
                empty_batch = False
                yield MessageProperties(
                    id=msg.message_id,
                    subject=msg.subject,
                )
                seq_num = msg.sequence_number + 1

            if empty_batch:
                break

    def iter_active_messages(self) -> Iterator[MessageProperties]:
        yield from self._iter_sub_queue_messages(sub_queue=None)

    def list_active_messages(self) -> list[MessageProperties]:
        return list(self.iter_active_messages())

    def iter_deadletter_messages(self) -> Iterator[MessageProperties]:
        yield from self._iter_sub_queue_messages(
            sub_queue=ServiceBusSubQueue.DEAD_LETTER
        )

    def list_deadletter_messages(self) -> list[MessageProperties]:
        return list(self.iter_deadletter_messages())



from typing import Iterator

from textual.widget import Widget

from azure_explorer.managers import StorageAccountManager
from azure_explorer.widgets.components.azure.container import BlobFolderBrowser
from azure_explorer.widgets.components.windows import Browser, Selector


class DataObjectTypeSelector(Selector):
    def __init__(self, storage_account_manager: StorageAccountManager):
        self.storage_account_manager = storage_account_manager
        super().__init__()

    def iter_options(self):
        yield ("container", "Containers")
        yield ("table", "Tables")

    def get_option_widget(self, id_):
        if id_ == "container":
            return ContainerExplorer(self.storage_account_manager)
        elif id_ == "table":
            return TableExplorer(self.storage_account_manager)
        else:
            raise ValueError(f"Unknown data object type: {id_}")


class ContainerExplorer(Browser):
    COLUMN_NAMES = ["Name"]

    def __init__(self, storage_account_manager: StorageAccountManager):
        self.storage_account_manager = storage_account_manager
        super().__init__()

    def check_valid(self) -> bool:
        self.storage_account_manager.check_access("container")

    def iter_rows(self) -> Iterator[tuple[str, tuple]]:
        for container_name in self.storage_account_manager.list_container_names():
            yield container_name, (f"📁 {container_name}",)

    def get_item_widget(self, item: str) -> Widget:
        container_manager = self.storage_account_manager.get_container_manager(
            container_name=item,
        )
        return BlobFolderBrowser(container_manager)


class TableExplorer(Browser):
    COLUMN_NAMES = ["Name"]

    def __init__(self, storage_account_manager: StorageAccountManager):
        self.storage_account_manager = storage_account_manager
        super().__init__()

    def check_valid(self) -> bool:
        self.storage_account_manager.check_access("table")

    def iter_rows(self) -> Iterator[tuple[str, tuple]]:
        for table_name in self.storage_account_manager.list_table_names():
            yield table_name, (f"📅 {table_name}",)



from abc import abstractmethod
from typing import Any, Iterator

from textual import events
from textual.app import ComposeResult
from textual.binding import Binding
from textual.widgets import DataTable, Input

from azure_explorer.widgets.components.windows.base import Window
from azure_explorer.widgets.utils import CachedIterator


class Browser(Window):

    BINDINGS = [
        Binding("home", "top_row", "Top row", show=False, priority=True),
        Binding("end", "bottom_row", "Bottom row", show=False, priority=True),
        Binding("up", "previous_row", "Previous row", show=False, priority=True),
        Binding("down", "next_row", "Next row", show=False, priority=True),
        Binding("pageup", "previous_page", "Previous page", show=True, priority=True),
        Binding("pagedown", "next_page", "Next page", show=True, priority=True),
        Binding("tab", "autocomplete", "Autocomplete", show=False, priority=True),
        Binding("ctrl+r", "refresh", "Refresh", show=True, priority=True),
    ]

    COLUMN_NAMES: list[str] = ...

    def __init__(self):
        self.search_filter = ""
        self.current_page = 0
        self.rows = None
        self.items_per_page = None
        self.filtered_rows = None

        super().__init__()

    def compose(self) -> ComposeResult:
        input_ = Input(placeholder="Search...")
        yield input_

        data_table = DataTable(cursor_type="row")
        data_table.can_focus = False
        yield data_table

    def check_action(self, action: str, parameters: tuple[object, ...]) -> bool | None:
        """Check if an action may run."""
        if action == "previous_page":
            if self.current_page == 0:
                return None
        elif action == "next_page":
            if not self.filtered_rows.done:
                return True
            num_rows = len(self.filtered_rows)
            last_page = (num_rows - 1) // self.items_per_page
            if self.current_page >= last_page:
                return None
        return True

    def action_previous_row(self):
        table = self.query_one(DataTable)
        table.action_cursor_up()

    def action_next_row(self):
        table = self.query_one(DataTable)
        table.action_cursor_down()

    def action_previous_page(self):
        self.current_page -= 1
        self.populate_table()

    def action_next_page(self):
        self.current_page += 1
        self.populate_table()

    def action_top_row(self):
        table = self.query_one(DataTable)
        table.action_scroll_top()

    def action_bottom_row(self):
        table = self.query_one(DataTable)
        table.action_scroll_bottom()

    def action_autocomplete(self):
        self.get_highlighted_item()
        input = self.query_one(Input)
        input.value = self.get_highlighted_item()

    def action_refresh(self):
        self.set_rows()
        self.set_filtered_rows()
        self.current_page = 0
        self.populate_table()

    def set_rows(self):
        self.rows = CachedIterator(self.iter_rows())

    def set_filtered_rows(self):
        self.filtered_rows = CachedIterator(self.iter_filtered_rows())

    def get_highlighted_item(self) -> Any:
        table = self.query_one(DataTable)
        if table.cursor_row is None:
            return None

        if not table.is_valid_coordinate(table.cursor_coordinate):
            return None

        selected_id = table.coordinate_to_cell_key(
            table.cursor_coordinate
        ).row_key.value
        return selected_id

    def _on_mount(self):
        self.items_per_page = self.app.size.height - 6

        self.set_rows()
        self.set_filtered_rows()

        table = self.query_one(DataTable)
        table.add_columns(*self.COLUMN_NAMES)

        self.populate_table()
        self.query_one(Input).focus()

    def on_data_table_row_selected(self, event: DataTable.RowSelected):
        selected_id = event.row_key.value

        try:
            widget = self.get_item_widget(selected_id)
        except Exception as exc:
            self.notify(str(exc), severity="error")
            return

        if widget is None:
            return

        self.open_child_window(widget)

    def on_input_changed(self, event: Input.Changed):
        self.search_filter = event.value.lower()
        self.set_filtered_rows()
        self.current_page = 0
        self.populate_table()

    def on_input_submitted(self):
        self.query_one(DataTable).action_select_cursor()

    def on_resize(self, event: events.Resize) -> None:
        self.items_per_page = self.app.size.height - 6
        self.current_page = 0
        self.populate_table()

    def populate_table(self):
        table = self.query_one(DataTable)
        table.clear()

        items_per_page = self.items_per_page
        start = self.current_page * items_per_page
        end = start + items_per_page

        for id_, item in self.filtered_rows[start:end]:
            table.add_row(*item, key=id_)

        self.refresh_bindings()

    def iter_filtered_rows(self) -> Iterator[tuple[str]]:
        search_filter = self.query_one(Input).value.lower()
        for id_, item in self.rows:
            valid = False
            for col in item:
                if search_filter in str(col).lower():
                    valid = True
            if valid:
                yield id_, item

    @abstractmethod
    def iter_rows(self) -> Iterator[tuple[str, tuple]]:
        ...

    @abstractmethod
    def get_item_widget(self, id_: str) -> "Window":
        ...



from typing import Iterator, Literal

from textual.widget import Widget

from azure_explorer.managers import (
    ServiceBusManager,
    TopicManager,
    TopicSubscriptionManager,
)
from azure_explorer.widgets.components.windows import Browser, Selector


class EntitySelector(Selector):
    def __init__(self, sb_manager: ServiceBusManager):
        self.sb_manager = sb_manager
        super().__init__()

    def iter_options(self):
        yield ("topic", "Topics")
        yield ("queue", "Queues")

    def get_option_widget(self, id_):
        if id_ == "topic":
            return TopicExplorer(self.sb_manager)


class TopicExplorer(Browser):
    COLUMN_NAMES = ["Name"]

    def __init__(self, sb_manager: ServiceBusManager):
        self.sb_manager = sb_manager
        super().__init__()

    def check_valid(self) -> bool:
        self.sb_manager.check_access()

    def iter_rows(self) -> Iterator[tuple[str, tuple]]:
        for topic in self.sb_manager.list_topics():
            yield topic.name, (topic.name,)

    def get_item_widget(self, item: str) -> Widget:
        topic_manager = self.sb_manager.get_topic_manager(
            topic_name=item,
        )
        return TopicSubscriptionExplorer(topic_manager)


class TopicSubscriptionExplorer(Browser):
    COLUMN_NAMES = ["Name", "Active Messages", "Dead-letter Messages"]

    def __init__(self, topic_manager: TopicManager):
        self.topic_manager = topic_manager
        super().__init__()

    def check_valid(self):
        return self.topic_manager.check_access()

    def iter_rows(self) -> Iterator[tuple[str, tuple]]:
        for sub in self.topic_manager.list_subscription():
            yield sub.name, (
                sub.name,
                sub.num_active_messages,
                sub.num_deadletter_messages,
            )

    def get_item_widget(self, id_: str) -> Widget:
        subscription_manager = self.topic_manager.get_subscription_manager(
            subscription_name=id_,
        )

        return SubQueueSelector(subscription_manager)


class SubQueueSelector(Selector):
    def __init__(self, subscription_manager: TopicSubscriptionManager):
        self.subscription_manager = subscription_manager
        super().__init__()

    def iter_options(self):
        yield ("active", "Active Messages")
        yield ("deadletter", "Dead-letter Messages")

    def get_option_widget(self, id_):
        return MessageSubjectExplorer(
            self.subscription_manager,
            id_,
        )


class MessageSubjectExplorer(Browser):
    COLUMN_NAMES = ["Subject"]

    def __init__(
        self,
        subscription_manager: TopicSubscriptionManager,
        sub_queue: Literal["active", "deadletter"],
    ):
        self.subscription_manager = subscription_manager
        self.sub_queue = sub_queue
        super().__init__()

    def check_valid(self):
        return self.subscription_manager.check_access()

    def iter_rows(self) -> Iterator[tuple[str, tuple]]:
        found_subjects = []

        if self.sub_queue == "deadletter":
            messages = self.subscription_manager.iter_deadletter_messages()
        elif self.sub_queue == "active":
            messages = self.subscription_manager.iter_active_messages()
        else:
            raise ValueError(f"Unknown sub-queue `{self.sub_queue}`")

        for message in messages:
            if message.subject in found_subjects:
                continue
            yield message.subject, (message.subject,)
            found_subjects.append(message.subject)

    def get_item_widget(self, id_: str) -> Widget:
        return MessageExplorer(
            self.subscription_manager,
            self.sub_queue,
            id_,
        )


class MessageExplorer(Browser):
    COLUMN_NAMES = ["Message ID", "Subject"]

    def __init__(
        self,
        subscription_manager: TopicSubscriptionManager,
        sub_queue: Literal["active", "deadletter"],
        subject: str,
    ):
        self.subscription_manager = subscription_manager
        self.sub_queue = sub_queue
        self.subject = subject
        super().__init__()

    def iter_rows(self) -> Iterator[tuple[str, tuple]]:
        if self.sub_queue == "deadletter":
            messages = self.subscription_manager.iter_deadletter_messages()
        else:
            messages = self.subscription_manager.iter_active_messages()

        for message in messages:
            if message.subject == self.subject:
                yield message.id, (message.id, message.subject)



from io import BytesIO
from pathlib import Path
from typing import Iterator

import pytz
from azure.core.credentials import TokenCredential
from azure.storage.blob import BlobPrefix, ContainerClient, StorageStreamDownloader

from azure_explorer.managers.container.base import ContainerManager
from azure_explorer.managers.models import FileProperties, FolderProperties

UTC = pytz.utc


class BlobStorageContainerManager(ContainerManager):
    def __init__(
        self,
        storage_account_name: str,
        container_name: str,
        credential: TokenCredential,
    ):
        self.storage_account_name = storage_account_name
        self.container_name = container_name
        self.credential = credential
        self.container_client = ContainerClient(
            f"https://{storage_account_name}.blob.core.windows.net",
            container_name=container_name,
            credential=credential,
        )

    def is_dir(self, path: Path | str) -> bool:
        if path == "":
            return True

        prefix = str(path).rstrip("/") + "/"

        try:
            first_blob = next(self.container_client.walk_blobs(name_starts_with=prefix))
        except StopIteration:
            return False

        if first_blob.name + "/" == prefix:
            return False

        return True

    def is_file(self, path: Path | str) -> bool:
        return self.container_client.get_blob_client(path).exists()

    def _iter_dir(self, path: Path | str, recursive: bool) -> Iterator[str]:
        if path == "":
            prefix = ""
        else:
            prefix = str(path).rstrip("/") + "/"

        for blob in self.container_client.walk_blobs(prefix, delimiter="/"):
            if not isinstance(blob, BlobPrefix):
                last_modified_time = blob.last_modified.replace(tzinfo=UTC)
                created_time = blob.creation_time.replace(tzinfo=UTC)
                yield FileProperties(
                    blob.name,
                    blob.size,
                    last_modified_time,
                    created_time,
                )
            else:
                if recursive:
                    yield from self._iter_dir(blob.name, recursive)
                else:
                    yield FolderProperties(
                        blob.name.rstrip("/"),
                        None,
                        None,
                        None,
                    )

    def _get_blob_stream_downloader(self, path: Path | str) -> StorageStreamDownloader:
        return self.container_client.download_blob(path)

    def _write_blob_data(
        self, path: Path | str, data: bytes | BytesIO, overwrite: bool = False
    ):
        self.container_client.upload_blob(path, data, overwrite=overwrite)

    def __repr__(self) -> str:
        return (
            "BlobStorageContainerManager("
            f"storage_account_name='{self.storage_account_name}', "
            f"container_name='{self.container_name}')"
        )



from .managers import (
    ContainerManager,
    StorageAccountManager,
    SubscriptionManager,
    TenantManager,
)
from .short_cuts import (
    get_container_manager,
    get_storage_account_manager,
    get_subscription_manager,
    get_tenant_manager,
    read_file,
)



import datetime as dt
from dataclasses import dataclass, field


@dataclass
class ServiceBusProperties:
    name: str


@dataclass
class SubscriptionProperties:
    id: str = field(repr=False)
    name: str


@dataclass
class StorageAccountProperties:
    name: str
    is_data_lake: bool


@dataclass
class TopicProperties:
    name: str


@dataclass
class TopicSubscriptionProperties:
    name: str
    num_active_messages: int = field(default=None)
    num_deadletter_messages: int = field(default=None)


@dataclass
class FileProperties:
    path: str
    size: float
    last_modified_time: dt.datetime
    created_time: dt.datetime


@dataclass
class FolderProperties:
    path: str
    size: float
    last_modified_time: dt.datetime
    created_time: dt.datetime


@dataclass
class MessageProperties:
    id: str
    subject: str



from typing import Iterator


class CachedIterator:
    def __init__(self, iterator: Iterator):
        """Wrap an iterable so that it functions like a list, but
        does so in a lazy manner. I.e. getting an item by a single index
        or index slice will consume the iterable up to that index,
        cache the items, and return the result. If the index has already
        been consumed, the cached items will be used.

        Args:
            iterable (_type_): Iterable
        """
        self.iterator = iterator
        self.current_index = -1
        self.cached_items = []
        self.done = False

    def __getitem__(self, index):
        if isinstance(index, int):
            if index < 0:
                raise IndexError("Negative indexing is not supported")
            max_index = index
        elif isinstance(index, slice):
            if index.start is not None and index.start < 0:
                raise IndexError("Negative indexing is not supported")
            if index.stop is None or index.stop < 0:
                raise IndexError("Negative indexing is not supported")
            max_index = index.stop - 1
        else:
            raise TypeError("Invalid index type")

        while self.current_index < max_index:
            try:
                item = next(self.iterator)
                self.cached_items.append(item)
                self.current_index += 1
            except StopIteration:
                self.done = True
                break

        return self.cached_items[index]

    def __len__(self):
        while True:
            try:
                item = next(self.iterator)
                self.cached_items.append(item)
                self.current_index += 1
            except StopIteration:
                self.done = True
                break

        return len(self.cached_items)



from typing import Iterator

from textual.widget import Widget

from azure_explorer.managers import SubscriptionManager
from azure_explorer.widgets.components.azure.service_bus import EntitySelector
from azure_explorer.widgets.components.azure.storage_account import (
    DataObjectTypeSelector,
)
from azure_explorer.widgets.components.windows import Browser, Selector


class ResourceTypeSelector(Selector):
    def __init__(self, sub_manager: SubscriptionManager):
        self.sub_manager = sub_manager
        super().__init__()

    def iter_options(self):
        yield ("storage_account", "Storage Accounts")
        yield ("data_lake", "Data Lakes")
        yield ("service_bus", "Service Buses")

    def get_option_widget(self, id_):
        if id_ == "storage_account":
            return StorageAccountBrowser(self.sub_manager, is_data_lake=False)
        elif id_ == "data_lake":
            return StorageAccountBrowser(self.sub_manager, is_data_lake=True)
        elif id_ == "service_bus":
            return ServiceBusBrowser(self.sub_manager)


class StorageAccountBrowser(Browser):
    COLUMN_NAMES = ["Name"]

    def __init__(self, subscription_manager: SubscriptionManager, is_data_lake: bool):
        self.subscription_manager = subscription_manager
        self.is_data_lake = is_data_lake
        super().__init__()

    def iter_rows(self) -> Iterator[tuple[str, tuple]]:
        for storage_account in self.subscription_manager.list_storage_accounts():
            if storage_account.is_data_lake == self.is_data_lake:
                yield storage_account.name, (storage_account.name,)

    def get_item_widget(self, id_: str) -> Widget:
        sa_manager = self.subscription_manager.get_storage_account_manager(
            storage_account_name=id_,
            is_data_lake=self.is_data_lake,
        )

        return DataObjectTypeSelector(sa_manager)


class ServiceBusBrowser(Browser):
    COLUMN_NAMES = ["Name"]

    def __init__(self, subscription_manager: SubscriptionManager):
        self.subscription_manager = subscription_manager
        super().__init__()

    def iter_rows(self) -> Iterator[tuple[str, tuple]]:
        for sb in self.subscription_manager.list_service_buses():
            yield sb.name, (sb.name,)

    def get_item_widget(self, item: str) -> Widget:
        sb_manager = self.subscription_manager.get_service_bus_manager(
            service_bus_name=item,
        )
        return EntitySelector(sb_manager)



from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.widgets import Footer, Header

from azure_explorer.config import Config
from azure_explorer.managers import TenantManager
from azure_explorer.widgets.components import SubscriptionBrowser


class AzureExplorerApp(App):

    BINDINGS = [
        Binding("ctrl+x", "quit", "Exit", show=True, priority=True),
    ]

    TITLE = "Azure Explorer"

    def compose(self) -> ComposeResult:
        yield Header()
        yield Footer()

    def on_mount(self):
        tenant_manager = TenantManager(credential=Config.credential)
        try:
            tenant_manager.check_access()
        except Exception as exc:
            self.notify(str(exc), severity="error")
            return

        sub_browser = SubscriptionBrowser(tenant_manager)
        self.mount(sub_browser)


def run():
    AzureExplorerApp().run()



from abc import abstractmethod
from typing import Iterator

from textual.app import ComposeResult
from textual.widgets import OptionList
from textual.widgets.option_list import Option

from azure_explorer.widgets.components.windows.base import Window


class Selector(Window):
    """Window representing a selector of options"""

    def compose(self) -> ComposeResult:
        options = []
        for id_, label in self.iter_options():
            options.append(Option(label, id=id_))

        option_list = OptionList(*options)

        option_list.focus()

        yield option_list

    def on_option_list_option_selected(self, event: OptionList.OptionSelected):
        selected_id = event.option_id

        try:
            widget = self.get_option_widget(selected_id)
        except Exception as exc:
            self.notify(str(exc), severity="error")
            return

        if widget is None:
            return

        self.open_child_window(widget)

    @abstractmethod
    def iter_options(self) -> Iterator[tuple[str, str]]:
        ...

    @abstractmethod
    def get_option_widget(self, id_: str) -> "Window":
        ...



from pathlib import Path
from typing import Iterator

from textual.binding import Binding
from textual.widget import Widget

from azure_explorer.config import Config
from azure_explorer.managers import ContainerManager
from azure_explorer.widgets.components.windows import Browser

READ_FILE_CODE_SNIPPET = """
import azure_explorer as ax

ax.read_file(
    storage_account_name="{self.container_manager.storage_account_name}",
    container_name="{self.container_manager.container_name}",
    path="{item}",
)
"""


class BlobFolderBrowser(Browser):
    COLUMN_NAMES = ["Name", "Size (MB)", "Last Modified", "Created Time"]

    BINDINGS = [
        Binding("ctrl+s", "download", "Download", show=True, priority=True),
        Binding("ctrl+c", "copy_code", "Copy code", show=True, priority=True),
    ]

    def __init__(self, container_manager: ContainerManager, folder: str = ""):
        self.container_manager = container_manager
        self.folder = folder
        super().__init__()

    def check_action(self, action: str, parameters: tuple[object, ...]) -> bool | None:
        """Check if an action may run."""
        if action == "copy_code":
            item = self.get_highlighted_item()
            if self.container_manager.is_dir(item):
                return None
        return super().check_action(action, parameters)

    def action_download(self):
        id_ = self.get_highlighted_item()

        path = f"{self.folder}/{id_}" if self.folder else id_

        download_folder = Path.cwd() / ".downloads" / path

        if self.container_manager.is_dir(path):
            self.container_manager.download_dir(path, download_folder)
        else:
            self.container_manager.download_file(path, download_folder)

    def action_copy_code(self):
        item = self.get_highlighted_item()
        code_snippet = READ_FILE_CODE_SNIPPET.format(
            self=self,
            item=item,
        )
        self.app.copy_to_clipboard(code_snippet)
        self.notify("Code snippet copied to clipboard!", severity="info", timeout=3)

    def iter_rows(self) -> Iterator[tuple[str, tuple]]:
        for subitem in self.container_manager.iter_dir(self.folder):
            if self.container_manager.is_dir(subitem.path):
                icon = "📁"
            else:
                icon = "📄"

            if subitem.path.endswith("/"):
                base_name = subitem.path
            else:
                base_name = subitem.path.split("/")[-1]

            if subitem.last_modified_time:
                last_modified_time_local = subitem.last_modified_time.astimezone(
                    Config.time_zone
                )
            else:
                last_modified_time_local = "-"

            if subitem.created_time:
                created_time_local = subitem.created_time.astimezone(Config.time_zone)
            else:
                created_time_local = "-"

            if subitem.size:
                size_MB = round(subitem.size / (1024 * 1024), 2)
            else:
                size_MB = "-"

            yield base_name, (
                f"{icon} {base_name}",
                size_MB,
                last_modified_time_local,
                created_time_local,
            )

    def get_item_widget(self, id_: str) -> Widget:
        path = f"{self.folder.rstrip('/')}/{id_}" if self.folder else id_

        if not self.container_manager.is_dir(path):
            return

        return BlobFolderBrowser(
            self.container_manager,
            path,
        )



from abc import ABC, abstractmethod
from io import BytesIO
from pathlib import Path
from typing import Iterator

from azure.core.credentials import TokenCredential
from azure.storage.blob import StorageStreamDownloader

from azure_explorer.managers.models import FileProperties, FolderProperties


class FolderNotFoundError(Exception):
    pass


class FileOrFolderNotFoundError(Exception):
    pass


class ContainerManager(ABC):
    @staticmethod
    def create(
        storage_account_name: str,
        container_name: str,
        is_data_lake: bool,
        credential: TokenCredential,
    ):
        from . import BlobStorageContainerManager, DataLakeContainerManager

        if is_data_lake:
            return DataLakeContainerManager(
                storage_account_name=storage_account_name,
                container_name=container_name,
                credential=credential,
            )
        else:
            return BlobStorageContainerManager(
                storage_account_name=storage_account_name,
                container_name=container_name,
                credential=credential,
            )

    @abstractmethod
    def _get_blob_stream_downloader(self, path: Path | str) -> StorageStreamDownloader:
        ...

    @abstractmethod
    def _write_blob_data(
        self, path: Path | str, data: bytes | BytesIO, overwrite: bool = False
    ):
        ...

    @abstractmethod
    def _iter_dir(
        self,
        path: Path | str,
        recursive: bool,
    ) -> Iterator[FileProperties | FolderProperties]:
        ...

    @abstractmethod
    def is_dir(self, path: Path | str) -> bool:
        ...

    @abstractmethod
    def is_file(self, path: Path | str) -> bool:
        ...

    def iter_dir(
        self, path: Path | str = "", recursive: bool = False
    ) -> Iterator[FileProperties | FolderProperties]:
        if not self.is_dir(path):
            raise FolderNotFoundError(path)

        yield from self._iter_dir(path, recursive)

    def list_dir(
        self, path: Path | str = "", recursive: bool = False
    ) -> list[FileProperties | FolderProperties]:
        return list(self.iter_dir(path, recursive))

    def read_file(self, path: Path | str) -> bytes:
        if not self.is_file(path):
            raise FileNotFoundError(path)

        blob_stream_downloader = self._get_blob_stream_downloader(path)

        return blob_stream_downloader.readall()

    def download_file(self, source: Path | str, destination: Path | str):
        destination = Path(destination)

        if not self.is_file(source):
            raise FileNotFoundError(source)

        destination.parent.mkdir(parents=True, exist_ok=True)

        blob_stream_downloader = self._get_blob_stream_downloader(source)

        with open(destination, "wb") as f:
            blob_stream_downloader.readinto(f)

    def download_dir(self, source: Path | str, destination: Path):
        destination = Path(destination)

        for subsource in self.iter_dir(source, recursive=True):
            subdestination = destination / Path(subsource.path)
            self.download_file(subsource.path, subdestination)

    def write_blob(self, data: bytes, path: Path | str, overwrite: bool = False):
        self._write_blob_data(path, data, overwrite=overwrite)

    def upload_file(self, source: Path | str, path: str, overwrite: bool = False):
        with open(source, "rb") as f:
            self._write_blob_data(path, f, overwrite=overwrite)

    def upload_dir(
        self, source: Path | str, destination: Path | str, overwrite: bool = False
    ):
        # NOTE: Will not create empty folders
        destination = Path(destination)
        source = Path(source)

        for subsource in source.rglob("*"):
            if not subsource.is_file():
                continue

            subdestination = str(destination / subsource.relative_to(source))

            self.upload_file(subsource, subdestination, overwrite)
